from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType

from .base import NotificationAction


@dataclass
class ValidateRuleAction(NotificationAction):
    """Action to validate a notification rule by rendering its template with test data."""

    rule_id: UUID
    notification_data: Mapping[str, Any]

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE

    @override
    def entity_id(self) -> str | None:
        return str(self.rule_id)


@dataclass
class ValidateRuleActionResult(BaseActionResult):
    """Result of validating a notification rule."""

    message: str

    @override
    def entity_id(self) -> str | None:
        return None



import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.common.data.session.types import CustomizedImageVisibilityScope
from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.session.types import SessionData
from ai.backend.manager.services.session.base import SessionAction


@dataclass
class ConvertSessionToImageAction(SessionAction):
    session_name: str
    owner_access_key: AccessKey
    image_name: str
    image_visibility: CustomizedImageVisibilityScope
    image_owner_id: uuid.UUID
    user_email: str
    max_customized_image_count: int

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class ConvertSessionToImageActionResult(BaseActionResult):
    task_id: uuid.UUID

    session_data: SessionData

    @override
    def entity_id(self) -> str | None:
        return str(self.session_data.id)



"""Domain Usage Bucket GQL types."""

from __future__ import annotations

from typing import Any, override

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.manager.api.gql.base import (
    DateFilter,
    OrderDirection,
    StringFilter,
)
from ai.backend.manager.api.gql.fair_share.types import ResourceSlotGQL
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy, StrawberryGQLContext
from ai.backend.manager.repositories.base import (
    QueryCondition,
    QueryOrder,
    combine_conditions_or,
    negate_conditions,
)
from ai.backend.manager.repositories.resource_usage_history.options import (
    DomainUsageBucketConditions,
    DomainUsageBucketOrders,
    ProjectUsageBucketConditions,
)
from ai.backend.manager.repositories.resource_usage_history.types import (
    DomainUsageBucketData,
)

from .common import UsageBucketMetadataGQL, UsageBucketOrderField
from .common_calculations import (
    calculate_average_daily_usage,
    calculate_usage_capacity_ratio,
)
from .project_usage import (
    ProjectUsageBucketConnection,
    ProjectUsageBucketFilter,
    ProjectUsageBucketOrderBy,
)


@strawberry.type(
    name="DomainUsageBucket",
    description=(
        "Added in 26.1.0. Domain-level usage bucket representing aggregated resource "
        "consumption for a specific time period. Usage buckets store historical data "
        "used to calculate fair share factors."
    ),
)
class DomainUsageBucketGQL(Node):
    """Domain-level usage bucket containing aggregated resource usage for a period."""

    id: NodeID[str]
    domain_name: str = strawberry.field(
        description="Name of the domain this usage bucket belongs to."
    )
    resource_group_name: str = strawberry.field(
        description="Name of the scaling group this usage was recorded in."
    )
    metadata: UsageBucketMetadataGQL = strawberry.field(
        description="Metadata about the usage measurement period and timestamps."
    )
    resource_usage: ResourceSlotGQL = strawberry.field(
        description=(
            "Aggregated resource usage during this period. "
            "Contains the total compute resources consumed by all workloads in this domain "
            "during the measurement period (cpu cores, memory bytes, accelerator usage)."
        )
    )
    capacity_snapshot: ResourceSlotGQL = strawberry.field(
        description=(
            "Snapshot of total available capacity in the scaling group at the end of this period. "
            "Used as a reference to calculate relative usage and fair share factors."
        )
    )

    @strawberry.field(  # type: ignore[misc]
        description=(
            "Added in 26.2.0. Average daily resource usage during this period. "
            "Calculated as resource_usage divided by bucket duration in days. "
            "For each resource type, this represents the average amount consumed per day. "
            "Units match the resource type (e.g., CPU cores, memory bytes)."
        )
    )
    def average_daily_usage(self) -> ResourceSlotGQL:
        return calculate_average_daily_usage(
            self.resource_usage,
            self.metadata.period_start,
            self.metadata.period_end,
        )

    @strawberry.field(  # type: ignore[misc]
        description=(
            "Added in 26.2.0. Usage ratio against total available capacity for each resource. "
            "Calculated as resource_usage divided by capacity_snapshot. "
            "Represents the fraction of total capacity consumed (resource-seconds / resource). "
            "The result is in seconds, where 86400 means full utilization for one day. "
            "Values can exceed this if usage exceeds capacity."
        )
    )
    def usage_capacity_ratio(self) -> ResourceSlotGQL:
        return calculate_usage_capacity_ratio(
            self.resource_usage,
            self.capacity_snapshot,
        )

    @classmethod
    def from_dataclass(cls, data: DomainUsageBucketData) -> DomainUsageBucketGQL:
        return cls(
            id=ID(str(data.id)),
            domain_name=data.domain_name,
            resource_group_name=data.resource_group,
            metadata=UsageBucketMetadataGQL(
                period_start=data.period_start,
                period_end=data.period_end,
                decay_unit_days=data.decay_unit_days,
                created_at=data.created_at,
                updated_at=data.updated_at,
            ),
            resource_usage=ResourceSlotGQL.from_resource_slot(data.resource_usage),
            capacity_snapshot=ResourceSlotGQL.from_resource_slot(data.capacity_snapshot),
        )

    @strawberry.field(  # type: ignore[misc]
        description=(
            "Added in 26.1.0. Project usage buckets belonging to this domain. "
            "Returns paginated project-level usage history for all projects in this domain "
            "within the same scaling group."
        )
    )
    async def project_usage_buckets(
        self,
        info: Info[StrawberryGQLContext],
        filter: ProjectUsageBucketFilter | None = None,
        order_by: list[ProjectUsageBucketOrderBy] | None = None,
        before: str | None = None,
        after: str | None = None,
        first: int | None = None,
        last: int | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> ProjectUsageBucketConnection:
        from ai.backend.manager.api.gql.resource_usage.fetcher.project_usage import (
            fetch_project_usage_buckets,
        )

        return await fetch_project_usage_buckets(
            info=info,
            filter=filter,
            order_by=order_by,
            before=before,
            after=after,
            first=first,
            last=last,
            limit=limit,
            offset=offset,
            base_conditions=[
                ProjectUsageBucketConditions.by_domain_name(self.domain_name),
                ProjectUsageBucketConditions.by_resource_group(self.resource_group_name),
                ProjectUsageBucketConditions.by_period_start(self.metadata.period_start),
            ],
        )


DomainUsageBucketEdge = Edge[DomainUsageBucketGQL]


@strawberry.type(
    description=(
        "Added in 26.1.0. Paginated connection for domain usage bucket records. "
        "Provides relay-style cursor-based pagination for browsing historical usage data by time period. "
        "Use 'edges' to access individual records with cursor information, or 'nodes' for direct data access."
    )
)
class DomainUsageBucketConnection(Connection[DomainUsageBucketGQL]):
    count: int = strawberry.field(
        description="Total number of domain usage bucket records matching the query criteria."
    )

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count


@strawberry.input(
    name="DomainUsageBucketFilter",
    description=(
        "Added in 26.1.0. Filter input for querying domain usage bucket records. "
        "Usage buckets contain historical resource consumption data aggregated by time period. "
        "Multiple filters can be combined using AND, OR, and NOT logical operators."
    ),
)
class DomainUsageBucketFilter(GQLFilter):
    """Filter for domain usage buckets."""

    resource_group: StringFilter | None = strawberry.field(
        default=None,
        description=(
            "Filter by scaling group name. Scaling groups define where usage was recorded. "
            "Supports equals, contains, startsWith, and endsWith operations."
        ),
    )
    domain_name: StringFilter | None = strawberry.field(
        default=None,
        description=(
            "Filter by domain name. This filters usage buckets for a specific domain. "
            "Supports equals, contains, startsWith, and endsWith operations."
        ),
    )
    period_start: DateFilter | None = strawberry.field(
        default=None, description="Filter by usage measurement period start date."
    )
    period_end: DateFilter | None = strawberry.field(
        default=None, description="Filter by usage measurement period end date."
    )

    AND: list[DomainUsageBucketFilter] | None = strawberry.field(
        default=None,
        description="Combine multiple filters with AND logic. All conditions must match.",
    )
    OR: list[DomainUsageBucketFilter] | None = strawberry.field(
        default=None,
        description="Combine multiple filters with OR logic. At least one condition must match.",
    )
    NOT: list[DomainUsageBucketFilter] | None = strawberry.field(
        default=None,
        description="Negate the specified filters. Records matching these conditions will be excluded.",
    )

    @override
    def build_conditions(self) -> list[QueryCondition]:
        conditions: list[QueryCondition] = []

        if self.resource_group:
            sg_condition = self.resource_group.build_query_condition(
                contains_factory=lambda spec: DomainUsageBucketConditions.by_resource_group_contains(
                    spec.value
                ),
                equals_factory=lambda spec: DomainUsageBucketConditions.by_resource_group_equals(
                    spec.value
                ),
                starts_with_factory=lambda spec: DomainUsageBucketConditions.by_resource_group_starts_with(
                    spec.value
                ),
                ends_with_factory=lambda spec: DomainUsageBucketConditions.by_resource_group_ends_with(
                    spec.value
                ),
            )
            if sg_condition:
                conditions.append(sg_condition)

        if self.domain_name:
            dn_condition = self.domain_name.build_query_condition(
                contains_factory=lambda spec: DomainUsageBucketConditions.by_domain_name_contains(
                    spec.value
                ),
                equals_factory=lambda spec: DomainUsageBucketConditions.by_domain_name_equals(
                    spec.value
                ),
                starts_with_factory=lambda spec: DomainUsageBucketConditions.by_domain_name_starts_with(
                    spec.value
                ),
                ends_with_factory=lambda spec: DomainUsageBucketConditions.by_domain_name_ends_with(
                    spec.value
                ),
            )
            if dn_condition:
                conditions.append(dn_condition)

        if self.period_start:
            ps_condition = self.period_start.build_query_condition(
                before_factory=DomainUsageBucketConditions.by_period_start_before,
                after_factory=DomainUsageBucketConditions.by_period_start_after,
                equals_factory=DomainUsageBucketConditions.by_period_start,
                not_equals_factory=DomainUsageBucketConditions.by_period_start_not_equals,
            )
            if ps_condition:
                conditions.append(ps_condition)

        if self.period_end:
            pe_condition = self.period_end.build_query_condition(
                before_factory=DomainUsageBucketConditions.by_period_end_before,
                after_factory=DomainUsageBucketConditions.by_period_end_after,
                equals_factory=DomainUsageBucketConditions.by_period_end,
                not_equals_factory=DomainUsageBucketConditions.by_period_end_not_equals,
            )
            if pe_condition:
                conditions.append(pe_condition)

        if self.AND:
            for sub_filter in self.AND:
                conditions.extend(sub_filter.build_conditions())

        if self.OR:
            or_conditions: list[QueryCondition] = []
            for sub_filter in self.OR:
                or_conditions.extend(sub_filter.build_conditions())
            if or_conditions:
                conditions.append(combine_conditions_or(or_conditions))

        if self.NOT:
            not_conditions: list[QueryCondition] = []
            for sub_filter in self.NOT:
                not_conditions.extend(sub_filter.build_conditions())
            if not_conditions:
                conditions.append(negate_conditions(not_conditions))

        return conditions


@strawberry.input(
    name="DomainUsageBucketOrderBy",
    description=(
        "Added in 26.1.0. Specifies ordering for domain usage bucket query results. "
        "Combine field selection with direction to sort results. "
        "Default direction is DESC (most recent first)."
    ),
)
class DomainUsageBucketOrderBy(GQLOrderBy):
    """OrderBy for domain usage buckets."""

    field: UsageBucketOrderField = strawberry.field(
        description="The field to order by. See UsageBucketOrderField for available options."
    )
    direction: OrderDirection = strawberry.field(
        default=OrderDirection.DESC,
        description=(
            "Sort direction. ASC for chronological order (oldest first), "
            "DESC for reverse chronological order (most recent first)."
        ),
    )

    @override
    def to_query_order(self) -> QueryOrder:
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case UsageBucketOrderField.PERIOD_START:
                return DomainUsageBucketOrders.by_period_start(ascending)


__all__ = [
    "DomainUsageBucketGQL",
    "DomainUsageBucketConnection",
    "DomainUsageBucketEdge",
    "DomainUsageBucketFilter",
    "DomainUsageBucketOrderBy",
]



"""
REST API handlers for fair share.
Provides search endpoints for fair share and usage bucket data.
"""

from __future__ import annotations

from collections.abc import Iterable
from http import HTTPStatus

import aiohttp_cors
from aiohttp import web

from ai.backend.common.api_handlers import APIResponse, BodyParam, PathParam, api_handler
from ai.backend.common.contexts.user import current_user
from ai.backend.common.dto.manager.fair_share import (
    BulkUpsertDomainFairShareWeightRequest,
    BulkUpsertDomainFairShareWeightResponse,
    BulkUpsertProjectFairShareWeightRequest,
    BulkUpsertProjectFairShareWeightResponse,
    BulkUpsertUserFairShareWeightRequest,
    BulkUpsertUserFairShareWeightResponse,
    DomainUsageBucketFilter,
    GetDomainFairSharePathParam,
    GetDomainFairShareResponse,
    GetProjectFairSharePathParam,
    GetProjectFairShareResponse,
    GetResourceGroupFairShareSpecPathParam,
    GetResourceGroupFairShareSpecResponse,
    GetUserFairSharePathParam,
    GetUserFairShareResponse,
    PaginationInfo,
    ProjectUsageBucketFilter,
    ResourceGroupFairShareSpecItemDTO,
    RGDomainFairSharePathParam,
    RGDomainFairShareSearchPathParam,
    RGDomainUsageBucketSearchPathParam,
    RGProjectFairSharePathParam,
    RGProjectFairShareSearchPathParam,
    RGProjectUsageBucketSearchPathParam,
    RGUserFairSharePathParam,
    RGUserFairShareSearchPathParam,
    RGUserUsageBucketSearchPathParam,
    SearchDomainFairSharesRequest,
    SearchDomainFairSharesResponse,
    SearchDomainUsageBucketsRequest,
    SearchDomainUsageBucketsResponse,
    SearchProjectFairSharesRequest,
    SearchProjectFairSharesResponse,
    SearchProjectUsageBucketsRequest,
    SearchProjectUsageBucketsResponse,
    SearchResourceGroupFairShareSpecsResponse,
    SearchUserFairSharesRequest,
    SearchUserFairSharesResponse,
    SearchUserUsageBucketsRequest,
    SearchUserUsageBucketsResponse,
    UpdateResourceGroupFairShareSpecPathParam,
    UpdateResourceGroupFairShareSpecRequest,
    UpdateResourceGroupFairShareSpecResponse,
    UpsertDomainFairShareWeightPathParam,
    UpsertDomainFairShareWeightRequest,
    UpsertDomainFairShareWeightResponse,
    UpsertProjectFairShareWeightPathParam,
    UpsertProjectFairShareWeightRequest,
    UpsertProjectFairShareWeightResponse,
    UpsertUserFairShareWeightPathParam,
    UpsertUserFairShareWeightRequest,
    UpsertUserFairShareWeightResponse,
    UserUsageBucketFilter,
)
from ai.backend.common.dto.manager.query import StringFilter, UUIDFilter
from ai.backend.manager.api.auth import auth_required_for_method
from ai.backend.manager.api.gql.base import StringMatchSpec
from ai.backend.manager.api.types import CORSOptions, WebMiddleware
from ai.backend.manager.dto.context import ProcessorsCtx
from ai.backend.manager.repositories.base import (
    BatchQuerier,
    NoPagination,
)
from ai.backend.manager.repositories.fair_share.types import (
    DomainFairShareSearchScope,
    ProjectFairShareSearchScope,
    UserFairShareSearchScope,
)
from ai.backend.manager.repositories.scaling_group.options import (
    ScalingGroupConditions,
)
from ai.backend.manager.services.fair_share.actions import (
    BulkUpsertDomainFairShareWeightAction,
    BulkUpsertProjectFairShareWeightAction,
    BulkUpsertUserFairShareWeightAction,
    DomainWeightInput,
    GetDomainFairShareAction,
    GetProjectFairShareAction,
    GetUserFairShareAction,
    ProjectWeightInput,
    SearchDomainFairSharesAction,
    SearchProjectFairSharesAction,
    SearchRGDomainFairSharesAction,
    SearchRGProjectFairSharesAction,
    SearchRGUserFairSharesAction,
    SearchUserFairSharesAction,
    UpsertDomainFairShareWeightAction,
    UpsertProjectFairShareWeightAction,
    UpsertUserFairShareWeightAction,
    UserWeightInput,
)
from ai.backend.manager.services.resource_usage.actions import (
    SearchDomainUsageBucketsAction,
    SearchProjectUsageBucketsAction,
    SearchUserUsageBucketsAction,
)
from ai.backend.manager.services.scaling_group.actions.list_scaling_groups import (
    SearchScalingGroupsAction,
)
from ai.backend.manager.services.scaling_group.actions.update_fair_share_spec import (
    ResourceWeightInput,
    UpdateFairShareSpecAction,
)

from .adapter import FairShareAdapter

__all__ = ("create_app",)


class FairShareAPIHandler:
    """REST API handler class for fair share operations."""

    def __init__(self) -> None:
        self._adapter = FairShareAdapter()

    def _check_superadmin(self) -> None:
        """Check if the current user is a superadmin."""
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can access fair share data.")

    # Domain Fair Share

    @auth_required_for_method
    @api_handler
    async def get_domain_fair_share(
        self,
        path: PathParam[GetDomainFairSharePathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get a single domain fair share."""
        self._check_superadmin()
        processors = processors_ctx.processors

        action_result = await processors.fair_share.get_domain_fair_share.wait_for_complete(
            GetDomainFairShareAction(
                resource_group=path.parsed.resource_group, domain_name=path.parsed.domain_name
            )
        )

        item = None
        if action_result.data is not None:
            item = self._adapter.convert_domain_fair_share_to_dto(action_result.data)

        resp = GetDomainFairShareResponse(item=item)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def search_domain_fair_shares(
        self,
        body: BodyParam[SearchDomainFairSharesRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search domain fair shares."""
        self._check_superadmin()
        processors = processors_ctx.processors

        querier = self._adapter.build_domain_fair_share_querier(body.parsed)

        action_result = await processors.fair_share.search_domain_fair_shares.wait_for_complete(
            SearchDomainFairSharesAction(
                pagination=querier.pagination,
                conditions=querier.conditions,
                orders=querier.orders,
            )
        )

        resp = SearchDomainFairSharesResponse(
            items=[
                self._adapter.convert_domain_fair_share_to_dto(fs) for fs in action_result.items
            ],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Project Fair Share

    @auth_required_for_method
    @api_handler
    async def get_project_fair_share(
        self,
        path: PathParam[GetProjectFairSharePathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get a single project fair share."""
        self._check_superadmin()
        processors = processors_ctx.processors

        action_result = await processors.fair_share.get_project_fair_share.wait_for_complete(
            GetProjectFairShareAction(
                resource_group=path.parsed.resource_group, project_id=path.parsed.project_id
            )
        )

        item = None
        if action_result.data is not None:
            item = self._adapter.convert_project_fair_share_to_dto(action_result.data)

        resp = GetProjectFairShareResponse(item=item)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def search_project_fair_shares(
        self,
        body: BodyParam[SearchProjectFairSharesRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search project fair shares."""
        self._check_superadmin()
        processors = processors_ctx.processors

        querier = self._adapter.build_project_fair_share_querier(body.parsed)

        action_result = await processors.fair_share.search_project_fair_shares.wait_for_complete(
            SearchProjectFairSharesAction(
                pagination=querier.pagination,
                conditions=querier.conditions,
                orders=querier.orders,
            )
        )

        resp = SearchProjectFairSharesResponse(
            items=[
                self._adapter.convert_project_fair_share_to_dto(fs) for fs in action_result.items
            ],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # User Fair Share

    @auth_required_for_method
    @api_handler
    async def get_user_fair_share(
        self,
        path: PathParam[GetUserFairSharePathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get a single user fair share."""
        self._check_superadmin()
        processors = processors_ctx.processors

        action_result = await processors.fair_share.get_user_fair_share.wait_for_complete(
            GetUserFairShareAction(
                resource_group=path.parsed.resource_group,
                project_id=path.parsed.project_id,
                user_uuid=path.parsed.user_uuid,
            )
        )

        item = None
        if action_result.data is not None:
            item = self._adapter.convert_user_fair_share_to_dto(action_result.data)

        resp = GetUserFairShareResponse(item=item)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def search_user_fair_shares(
        self,
        body: BodyParam[SearchUserFairSharesRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search user fair shares."""
        self._check_superadmin()
        processors = processors_ctx.processors

        querier = self._adapter.build_user_fair_share_querier(body.parsed)

        action_result = await processors.fair_share.search_user_fair_shares.wait_for_complete(
            SearchUserFairSharesAction(
                pagination=querier.pagination,
                conditions=querier.conditions,
                orders=querier.orders,
            )
        )

        resp = SearchUserFairSharesResponse(
            items=[self._adapter.convert_user_fair_share_to_dto(fs) for fs in action_result.items],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Domain Usage Bucket

    @auth_required_for_method
    @api_handler
    async def search_domain_usage_buckets(
        self,
        body: BodyParam[SearchDomainUsageBucketsRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search domain usage buckets."""
        self._check_superadmin()
        processors = processors_ctx.processors

        querier = self._adapter.build_domain_usage_bucket_querier(body.parsed)

        action_result = (
            await processors.resource_usage.search_domain_usage_buckets.wait_for_complete(
                SearchDomainUsageBucketsAction(
                    pagination=querier.pagination,
                    conditions=querier.conditions,
                    orders=querier.orders,
                )
            )
        )

        resp = SearchDomainUsageBucketsResponse(
            items=[
                self._adapter.convert_domain_usage_bucket_to_dto(b) for b in action_result.items
            ],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Project Usage Bucket

    @auth_required_for_method
    @api_handler
    async def search_project_usage_buckets(
        self,
        body: BodyParam[SearchProjectUsageBucketsRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search project usage buckets."""
        self._check_superadmin()
        processors = processors_ctx.processors

        querier = self._adapter.build_project_usage_bucket_querier(body.parsed)

        action_result = (
            await processors.resource_usage.search_project_usage_buckets.wait_for_complete(
                SearchProjectUsageBucketsAction(
                    pagination=querier.pagination,
                    conditions=querier.conditions,
                    orders=querier.orders,
                )
            )
        )

        resp = SearchProjectUsageBucketsResponse(
            items=[
                self._adapter.convert_project_usage_bucket_to_dto(b) for b in action_result.items
            ],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # User Usage Bucket

    @auth_required_for_method
    @api_handler
    async def search_user_usage_buckets(
        self,
        body: BodyParam[SearchUserUsageBucketsRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search user usage buckets."""
        self._check_superadmin()
        processors = processors_ctx.processors

        querier = self._adapter.build_user_usage_bucket_querier(body.parsed)

        action_result = await processors.resource_usage.search_user_usage_buckets.wait_for_complete(
            SearchUserUsageBucketsAction(
                pagination=querier.pagination,
                conditions=querier.conditions,
                orders=querier.orders,
            )
        )

        resp = SearchUserUsageBucketsResponse(
            items=[self._adapter.convert_user_usage_bucket_to_dto(b) for b in action_result.items],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # RG-Scoped Domain Usage Bucket

    @auth_required_for_method
    @api_handler
    async def rg_search_domain_usage_buckets(
        self,
        path: PathParam[RGDomainUsageBucketSearchPathParam],
        body: BodyParam[SearchDomainUsageBucketsRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search domain usage buckets within resource group scope."""
        processors = processors_ctx.processors

        # Filter에 resource_group 강제 주입
        filter = body.parsed.filter
        if filter is None:
            filter = DomainUsageBucketFilter(
                resource_group=StringFilter(equals=path.parsed.resource_group)
            )
        else:
            if filter.resource_group is None:
                filter.resource_group = StringFilter(equals=path.parsed.resource_group)
            elif filter.resource_group.equals != path.parsed.resource_group:
                raise web.HTTPBadRequest(reason="Filter resource_group must match path parameter")

        modified_request = SearchDomainUsageBucketsRequest(
            filter=filter,
            order=body.parsed.order,
            limit=body.parsed.limit,
            offset=body.parsed.offset,
        )

        querier = self._adapter.build_domain_usage_bucket_querier(modified_request)

        action_result = (
            await processors.resource_usage.search_domain_usage_buckets.wait_for_complete(
                SearchDomainUsageBucketsAction(
                    pagination=querier.pagination,
                    conditions=querier.conditions,
                    orders=querier.orders,
                )
            )
        )

        resp = SearchDomainUsageBucketsResponse(
            items=[
                self._adapter.convert_domain_usage_bucket_to_dto(b) for b in action_result.items
            ],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # RG-Scoped Project Usage Bucket

    @auth_required_for_method
    @api_handler
    async def rg_search_project_usage_buckets(
        self,
        path: PathParam[RGProjectUsageBucketSearchPathParam],
        body: BodyParam[SearchProjectUsageBucketsRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search project usage buckets within resource group and domain scope."""
        processors = processors_ctx.processors

        # Filter에 resource_group, domain_name 강제 주입
        filter = body.parsed.filter
        if filter is None:
            filter = ProjectUsageBucketFilter(
                resource_group=StringFilter(equals=path.parsed.resource_group),
                domain_name=StringFilter(equals=path.parsed.domain_name),
            )
        else:
            if filter.resource_group is None:
                filter.resource_group = StringFilter(equals=path.parsed.resource_group)
            elif filter.resource_group.equals != path.parsed.resource_group:
                raise web.HTTPBadRequest(reason="Filter resource_group must match path parameter")

            if filter.domain_name is None:
                filter.domain_name = StringFilter(equals=path.parsed.domain_name)
            elif filter.domain_name.equals != path.parsed.domain_name:
                raise web.HTTPBadRequest(reason="Filter domain_name must match path parameter")

        modified_request = SearchProjectUsageBucketsRequest(
            filter=filter,
            order=body.parsed.order,
            limit=body.parsed.limit,
            offset=body.parsed.offset,
        )

        querier = self._adapter.build_project_usage_bucket_querier(modified_request)

        action_result = (
            await processors.resource_usage.search_project_usage_buckets.wait_for_complete(
                SearchProjectUsageBucketsAction(
                    pagination=querier.pagination,
                    conditions=querier.conditions,
                    orders=querier.orders,
                )
            )
        )

        resp = SearchProjectUsageBucketsResponse(
            items=[
                self._adapter.convert_project_usage_bucket_to_dto(b) for b in action_result.items
            ],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # RG-Scoped User Usage Bucket

    @auth_required_for_method
    @api_handler
    async def rg_search_user_usage_buckets(
        self,
        path: PathParam[RGUserUsageBucketSearchPathParam],
        body: BodyParam[SearchUserUsageBucketsRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search user usage buckets within resource group, domain, and project scope."""
        processors = processors_ctx.processors

        # Filter에 resource_group, domain_name, project_id 강제 주입
        filter = body.parsed.filter
        if filter is None:
            filter = UserUsageBucketFilter(
                resource_group=StringFilter(equals=path.parsed.resource_group),
                domain_name=StringFilter(equals=path.parsed.domain_name),
                project_id=UUIDFilter(equals=path.parsed.project_id),
            )
        else:
            if filter.resource_group is None:
                filter.resource_group = StringFilter(equals=path.parsed.resource_group)
            elif filter.resource_group.equals != path.parsed.resource_group:
                raise web.HTTPBadRequest(reason="Filter resource_group must match path parameter")

            if filter.domain_name is None:
                filter.domain_name = StringFilter(equals=path.parsed.domain_name)
            elif filter.domain_name.equals != path.parsed.domain_name:
                raise web.HTTPBadRequest(reason="Filter domain_name must match path parameter")

            if filter.project_id is None:
                filter.project_id = UUIDFilter(equals=path.parsed.project_id)
            elif filter.project_id.equals != path.parsed.project_id:
                raise web.HTTPBadRequest(reason="Filter project_id must match path parameter")

        modified_request = SearchUserUsageBucketsRequest(
            filter=filter,
            order=body.parsed.order,
            limit=body.parsed.limit,
            offset=body.parsed.offset,
        )

        querier = self._adapter.build_user_usage_bucket_querier(modified_request)

        action_result = await processors.resource_usage.search_user_usage_buckets.wait_for_complete(
            SearchUserUsageBucketsAction(
                pagination=querier.pagination,
                conditions=querier.conditions,
                orders=querier.orders,
            )
        )

        resp = SearchUserUsageBucketsResponse(
            items=[self._adapter.convert_user_usage_bucket_to_dto(b) for b in action_result.items],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # RG-Scoped Domain Fair Share

    @auth_required_for_method
    @api_handler
    async def rg_get_domain_fair_share(
        self,
        path: PathParam[RGDomainFairSharePathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get a single domain fair share within RG scope."""
        processors = processors_ctx.processors

        action_result = await processors.fair_share.get_domain_fair_share.wait_for_complete(
            GetDomainFairShareAction(
                resource_group=path.parsed.resource_group, domain_name=path.parsed.domain_name
            )
        )

        item = None
        if action_result.data is not None:
            item = self._adapter.convert_domain_fair_share_to_dto(action_result.data)

        resp = GetDomainFairShareResponse(item=item)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def rg_search_domain_fair_shares(
        self,
        path: PathParam[RGDomainFairShareSearchPathParam],
        body: BodyParam[SearchDomainFairSharesRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search domain fair shares within RG scope."""
        processors = processors_ctx.processors

        querier = self._adapter.build_domain_fair_share_querier(body.parsed)
        scope = DomainFairShareSearchScope(resource_group=path.parsed.resource_group)

        action_result = await processors.fair_share.search_rg_domain_fair_shares.wait_for_complete(
            SearchRGDomainFairSharesAction(
                scope=scope,
                querier=querier,
            )
        )

        resp = SearchDomainFairSharesResponse(
            items=[
                self._adapter.convert_domain_fair_share_to_dto(fs) for fs in action_result.items
            ],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # RG-Scoped Project Fair Share

    @auth_required_for_method
    @api_handler
    async def rg_get_project_fair_share(
        self,
        path: PathParam[RGProjectFairSharePathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get a single project fair share within RG scope."""
        processors = processors_ctx.processors

        action_result = await processors.fair_share.get_project_fair_share.wait_for_complete(
            GetProjectFairShareAction(
                resource_group=path.parsed.resource_group, project_id=path.parsed.project_id
            )
        )

        item = None
        if action_result.data is not None:
            item = self._adapter.convert_project_fair_share_to_dto(action_result.data)

        resp = GetProjectFairShareResponse(item=item)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def rg_search_project_fair_shares(
        self,
        path: PathParam[RGProjectFairShareSearchPathParam],
        body: BodyParam[SearchProjectFairSharesRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search project fair shares within RG scope."""
        processors = processors_ctx.processors

        querier = self._adapter.build_project_fair_share_querier(body.parsed)
        scope = ProjectFairShareSearchScope(
            resource_group=path.parsed.resource_group,
            domain_name=path.parsed.domain_name,
        )

        action_result = await processors.fair_share.search_rg_project_fair_shares.wait_for_complete(
            SearchRGProjectFairSharesAction(
                scope=scope,
                querier=querier,
            )
        )

        resp = SearchProjectFairSharesResponse(
            items=[
                self._adapter.convert_project_fair_share_to_dto(fs) for fs in action_result.items
            ],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # RG-Scoped User Fair Share

    @auth_required_for_method
    @api_handler
    async def rg_get_user_fair_share(
        self,
        path: PathParam[RGUserFairSharePathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get a single user fair share within RG scope."""
        processors = processors_ctx.processors

        action_result = await processors.fair_share.get_user_fair_share.wait_for_complete(
            GetUserFairShareAction(
                resource_group=path.parsed.resource_group,
                project_id=path.parsed.project_id,
                user_uuid=path.parsed.user_uuid,
            )
        )

        item = None
        if action_result.data is not None:
            item = self._adapter.convert_user_fair_share_to_dto(action_result.data)

        resp = GetUserFairShareResponse(item=item)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def rg_search_user_fair_shares(
        self,
        path: PathParam[RGUserFairShareSearchPathParam],
        body: BodyParam[SearchUserFairSharesRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search user fair shares within RG scope."""
        processors = processors_ctx.processors

        querier = self._adapter.build_user_fair_share_querier(body.parsed)
        scope = UserFairShareSearchScope(
            resource_group=path.parsed.resource_group,
            domain_name=path.parsed.domain_name,
            project_id=path.parsed.project_id,
        )

        action_result = await processors.fair_share.search_rg_user_fair_shares.wait_for_complete(
            SearchRGUserFairSharesAction(
                scope=scope,
                querier=querier,
            )
        )

        resp = SearchUserFairSharesResponse(
            items=[self._adapter.convert_user_fair_share_to_dto(fs) for fs in action_result.items],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Upsert Domain Fair Share Weight

    @auth_required_for_method
    @api_handler
    async def upsert_domain_fair_share_weight(
        self,
        path: PathParam[UpsertDomainFairShareWeightPathParam],
        body: BodyParam[UpsertDomainFairShareWeightRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Upsert domain fair share weight."""
        self._check_superadmin()
        processors = processors_ctx.processors

        action_result = (
            await processors.fair_share.upsert_domain_fair_share_weight.wait_for_complete(
                UpsertDomainFairShareWeightAction(
                    resource_group=path.parsed.resource_group,
                    domain_name=path.parsed.domain_name,
                    weight=body.parsed.weight,
                )
            )
        )

        item = self._adapter.convert_domain_fair_share_to_dto(action_result.data)
        resp = UpsertDomainFairShareWeightResponse(item=item)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Upsert Project Fair Share Weight

    @auth_required_for_method
    @api_handler
    async def upsert_project_fair_share_weight(
        self,
        path: PathParam[UpsertProjectFairShareWeightPathParam],
        body: BodyParam[UpsertProjectFairShareWeightRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Upsert project fair share weight."""
        self._check_superadmin()
        processors = processors_ctx.processors

        action_result = (
            await processors.fair_share.upsert_project_fair_share_weight.wait_for_complete(
                UpsertProjectFairShareWeightAction(
                    resource_group=path.parsed.resource_group,
                    project_id=path.parsed.project_id,
                    domain_name=body.parsed.domain_name,
                    weight=body.parsed.weight,
                )
            )
        )

        item = self._adapter.convert_project_fair_share_to_dto(action_result.data)
        resp = UpsertProjectFairShareWeightResponse(item=item)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Upsert User Fair Share Weight

    @auth_required_for_method
    @api_handler
    async def upsert_user_fair_share_weight(
        self,
        path: PathParam[UpsertUserFairShareWeightPathParam],
        body: BodyParam[UpsertUserFairShareWeightRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Upsert user fair share weight."""
        self._check_superadmin()
        processors = processors_ctx.processors

        action_result = await processors.fair_share.upsert_user_fair_share_weight.wait_for_complete(
            UpsertUserFairShareWeightAction(
                resource_group=path.parsed.resource_group,
                project_id=path.parsed.project_id,
                user_uuid=path.parsed.user_uuid,
                domain_name=body.parsed.domain_name,
                weight=body.parsed.weight,
            )
        )

        item = self._adapter.convert_user_fair_share_to_dto(action_result.data)
        resp = UpsertUserFairShareWeightResponse(item=item)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Bulk Upsert Domain Fair Share Weight

    @auth_required_for_method
    @api_handler
    async def bulk_upsert_domain_fair_share_weight(
        self,
        body: BodyParam[BulkUpsertDomainFairShareWeightRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Bulk upsert domain fair share weights."""
        self._check_superadmin()
        processors = processors_ctx.processors

        # Convert DTO inputs to action inputs
        inputs = [
            DomainWeightInput(
                domain_name=entry.domain_name,
                weight=entry.weight,
            )
            for entry in body.parsed.inputs
        ]

        action_result = (
            await processors.fair_share.bulk_upsert_domain_fair_share_weight.wait_for_complete(
                BulkUpsertDomainFairShareWeightAction(
                    resource_group=body.parsed.resource_group,
                    inputs=inputs,
                )
            )
        )

        resp = BulkUpsertDomainFairShareWeightResponse(upserted_count=action_result.upserted_count)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Bulk Upsert Project Fair Share Weight

    @auth_required_for_method
    @api_handler
    async def bulk_upsert_project_fair_share_weight(
        self,
        body: BodyParam[BulkUpsertProjectFairShareWeightRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Bulk upsert project fair share weights."""
        self._check_superadmin()
        processors = processors_ctx.processors

        # Convert DTO inputs to action inputs
        inputs = [
            ProjectWeightInput(
                project_id=entry.project_id,
                domain_name=entry.domain_name,
                weight=entry.weight,
            )
            for entry in body.parsed.inputs
        ]

        action_result = (
            await processors.fair_share.bulk_upsert_project_fair_share_weight.wait_for_complete(
                BulkUpsertProjectFairShareWeightAction(
                    resource_group=body.parsed.resource_group,
                    inputs=inputs,
                )
            )
        )

        resp = BulkUpsertProjectFairShareWeightResponse(upserted_count=action_result.upserted_count)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Bulk Upsert User Fair Share Weight

    @auth_required_for_method
    @api_handler
    async def bulk_upsert_user_fair_share_weight(
        self,
        body: BodyParam[BulkUpsertUserFairShareWeightRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Bulk upsert user fair share weights."""
        self._check_superadmin()
        processors = processors_ctx.processors

        # Convert DTO inputs to action inputs
        inputs = [
            UserWeightInput(
                user_uuid=entry.user_uuid,
                project_id=entry.project_id,
                domain_name=entry.domain_name,
                weight=entry.weight,
            )
            for entry in body.parsed.inputs
        ]

        action_result = (
            await processors.fair_share.bulk_upsert_user_fair_share_weight.wait_for_complete(
                BulkUpsertUserFairShareWeightAction(
                    resource_group=body.parsed.resource_group,
                    inputs=inputs,
                )
            )
        )

        resp = BulkUpsertUserFairShareWeightResponse(upserted_count=action_result.upserted_count)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Update Resource Group Fair Share Spec

    @auth_required_for_method
    @api_handler
    async def get_resource_group_fair_share_spec(
        self,
        path: PathParam[GetResourceGroupFairShareSpecPathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get resource group fair share spec."""
        self._check_superadmin()
        processors = processors_ctx.processors

        # Get scaling group by name
        name_spec = StringMatchSpec(
            value=path.parsed.resource_group,
            case_insensitive=False,
            negated=False,
        )
        querier = BatchQuerier(
            pagination=NoPagination(),
            conditions=[ScalingGroupConditions.by_name_equals(name_spec)],
        )
        search_result = await processors.scaling_group.search_scaling_groups.wait_for_complete(
            SearchScalingGroupsAction(querier=querier)
        )

        if not search_result.scaling_groups:
            raise web.HTTPNotFound(
                reason=f"Resource group '{path.parsed.resource_group}' not found"
            )

        scaling_group = search_result.scaling_groups[0]

        resp = GetResourceGroupFairShareSpecResponse(
            resource_group=scaling_group.name,
            fair_share_spec=self._adapter.convert_scaling_group_spec_to_dto(
                scaling_group.fair_share_spec
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def search_resource_group_fair_share_specs(
        self,
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search all resource groups with their fair share specs."""
        self._check_superadmin()
        processors = processors_ctx.processors

        # Get all scaling groups
        querier = BatchQuerier(
            pagination=NoPagination(),
            conditions=[],
        )
        search_result = await processors.scaling_group.search_scaling_groups.wait_for_complete(
            SearchScalingGroupsAction(querier=querier)
        )

        items = [
            ResourceGroupFairShareSpecItemDTO(
                resource_group=sg.name,
                fair_share_spec=self._adapter.convert_scaling_group_spec_to_dto(sg.fair_share_spec),
            )
            for sg in search_result.scaling_groups
        ]

        resp = SearchResourceGroupFairShareSpecsResponse(
            items=items,
            total_count=len(items),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def update_resource_group_fair_share_spec(
        self,
        path: PathParam[UpdateResourceGroupFairShareSpecPathParam],
        body: BodyParam[UpdateResourceGroupFairShareSpecRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Update resource group fair share spec with partial update and validation."""
        self._check_superadmin()
        processors = processors_ctx.processors

        # Convert request to action
        resource_weights = None
        if body.parsed.resource_weights is not None:
            resource_weights = [
                ResourceWeightInput(
                    resource_type=entry.resource_type,
                    weight=entry.weight,
                )
                for entry in body.parsed.resource_weights
            ]

        action = UpdateFairShareSpecAction(
            resource_group=path.parsed.resource_group,
            half_life_days=body.parsed.half_life_days,
            lookback_days=body.parsed.lookback_days,
            decay_unit_days=body.parsed.decay_unit_days,
            default_weight=body.parsed.default_weight,
            resource_weights=resource_weights,
        )

        result = await processors.scaling_group.update_fair_share_spec.wait_for_complete(action)

        resp = UpdateResourceGroupFairShareSpecResponse(
            resource_group=result.scaling_group.name,
            fair_share_spec=self._adapter.convert_scaling_group_spec_to_dto(
                result.scaling_group.fair_share_spec
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    """Create aiohttp application for fair share API endpoints."""
    app = web.Application()
    app["api_versions"] = (4, 5)
    app["prefix"] = "fair-share"

    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    api_handler = FairShareAPIHandler()

    # Domain fair share routes
    cors.add(
        app.router.add_route(
            "GET",
            "/domains/{resource_group}/{domain_name}",
            api_handler.get_domain_fair_share,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/domains/search",
            api_handler.search_domain_fair_shares,
        )
    )

    # Project fair share routes
    cors.add(
        app.router.add_route(
            "GET",
            "/projects/{resource_group}/{project_id}",
            api_handler.get_project_fair_share,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/projects/search",
            api_handler.search_project_fair_shares,
        )
    )

    # User fair share routes
    cors.add(
        app.router.add_route(
            "GET",
            "/users/{resource_group}/{project_id}/{user_uuid}",
            api_handler.get_user_fair_share,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/users/search",
            api_handler.search_user_fair_shares,
        )
    )

    # Usage bucket routes
    cors.add(
        app.router.add_route(
            "POST",
            "/usage-buckets/domains/search",
            api_handler.search_domain_usage_buckets,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/usage-buckets/projects/search",
            api_handler.search_project_usage_buckets,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/usage-buckets/users/search",
            api_handler.search_user_usage_buckets,
        )
    )

    # RG-scoped usage bucket routes
    cors.add(
        app.router.add_route(
            "POST",
            "/rg/{resource_group}/usage-buckets/domains/search",
            api_handler.rg_search_domain_usage_buckets,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/rg/{resource_group}/domains/{domain_name}/usage-buckets/projects/search",
            api_handler.rg_search_project_usage_buckets,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/rg/{resource_group}/domains/{domain_name}/projects/{project_id}/usage-buckets/users/search",
            api_handler.rg_search_user_usage_buckets,
        )
    )

    # RG-scoped domain fair share routes
    cors.add(
        app.router.add_route(
            "GET",
            "/rg/{resource_group}/domains/{domain_name}",
            api_handler.rg_get_domain_fair_share,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/rg/{resource_group}/domains/search",
            api_handler.rg_search_domain_fair_shares,
        )
    )

    # RG-scoped project fair share routes
    cors.add(
        app.router.add_route(
            "GET",
            "/rg/{resource_group}/domains/{domain_name}/projects/{project_id}",
            api_handler.rg_get_project_fair_share,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/rg/{resource_group}/domains/{domain_name}/projects/search",
            api_handler.rg_search_project_fair_shares,
        )
    )

    # RG-scoped user fair share routes
    cors.add(
        app.router.add_route(
            "GET",
            "/rg/{resource_group}/domains/{domain_name}/projects/{project_id}/users/{user_uuid}",
            api_handler.rg_get_user_fair_share,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/rg/{resource_group}/domains/{domain_name}/projects/{project_id}/users/search",
            api_handler.rg_search_user_fair_shares,
        )
    )

    # Upsert weight routes
    cors.add(
        app.router.add_route(
            "PUT",
            "/domains/{resource_group}/{domain_name}/weight",
            api_handler.upsert_domain_fair_share_weight,
        )
    )
    cors.add(
        app.router.add_route(
            "PUT",
            "/projects/{resource_group}/{project_id}/weight",
            api_handler.upsert_project_fair_share_weight,
        )
    )
    cors.add(
        app.router.add_route(
            "PUT",
            "/users/{resource_group}/{project_id}/{user_uuid}/weight",
            api_handler.upsert_user_fair_share_weight,
        )
    )

    # Bulk upsert weight routes
    cors.add(
        app.router.add_route(
            "POST",
            "/domains/bulk-upsert-weight",
            api_handler.bulk_upsert_domain_fair_share_weight,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/projects/bulk-upsert-weight",
            api_handler.bulk_upsert_project_fair_share_weight,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/users/bulk-upsert-weight",
            api_handler.bulk_upsert_user_fair_share_weight,
        )
    )

    # Resource group spec routes
    cors.add(
        app.router.add_route(
            "GET",
            "/resource-groups/{resource_group}/spec",
            api_handler.get_resource_group_fair_share_spec,
        )
    )
    cors.add(
        app.router.add_route(
            "GET",
            "/resource-groups/specs",
            api_handler.search_resource_group_fair_share_specs,
        )
    )
    cors.add(
        app.router.add_route(
            "PATCH",
            "/resource-groups/{resource_group}/spec",
            api_handler.update_resource_group_fair_share_spec,
        )
    )

    return app, []



from __future__ import annotations

import uuid
from collections.abc import Collection

import sqlalchemy as sa

from ai.backend.common.data.notification import NotificationChannelType, NotificationRuleType
from ai.backend.manager.api.gql.base import StringMatchSpec
from ai.backend.manager.models.notification import NotificationChannelRow, NotificationRuleRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder


class NotificationChannelConditions:
    """Query conditions for notification channels."""

    @staticmethod
    def by_ids(channel_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return NotificationChannelRow.id.in_(channel_ids)

        return inner

    @staticmethod
    def by_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = NotificationChannelRow.name.ilike(f"%{spec.value}%")
            else:
                condition = NotificationChannelRow.name.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(NotificationChannelRow.name) == spec.value.lower()
            else:
                condition = NotificationChannelRow.name == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = NotificationChannelRow.name.ilike(f"{spec.value}%")
            else:
                condition = NotificationChannelRow.name.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = NotificationChannelRow.name.ilike(f"%{spec.value}")
            else:
                condition = NotificationChannelRow.name.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_channel_types(channel_types: list[NotificationChannelType]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return NotificationChannelRow.channel_type.in_(channel_types)

        return inner

    @staticmethod
    def by_enabled(enabled: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return NotificationChannelRow.enabled == enabled

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(NotificationChannelRow.created_at)
                .where(NotificationChannelRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return NotificationChannelRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(NotificationChannelRow.created_at)
                .where(NotificationChannelRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return NotificationChannelRow.created_at > subquery

        return inner


class NotificationChannelOrders:
    """Query orders for notification channels."""

    @staticmethod
    def name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return NotificationChannelRow.name.asc()
        return NotificationChannelRow.name.desc()

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return NotificationChannelRow.created_at.asc()
        return NotificationChannelRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return NotificationChannelRow.updated_at.asc()
        return NotificationChannelRow.updated_at.desc()


class NotificationRuleConditions:
    """Query conditions for notification rules."""

    @staticmethod
    def by_ids(rule_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return NotificationRuleRow.id.in_(rule_ids)

        return inner

    @staticmethod
    def by_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = NotificationRuleRow.name.ilike(f"%{spec.value}%")
            else:
                condition = NotificationRuleRow.name.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(NotificationRuleRow.name) == spec.value.lower()
            else:
                condition = NotificationRuleRow.name == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = NotificationRuleRow.name.ilike(f"{spec.value}%")
            else:
                condition = NotificationRuleRow.name.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = NotificationRuleRow.name.ilike(f"%{spec.value}")
            else:
                condition = NotificationRuleRow.name.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_rule_types(rule_types: list[NotificationRuleType]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return NotificationRuleRow.rule_type.in_(rule_types)

        return inner

    @staticmethod
    def by_enabled(enabled: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return NotificationRuleRow.enabled == enabled

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(NotificationRuleRow.created_at)
                .where(NotificationRuleRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return NotificationRuleRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(NotificationRuleRow.created_at)
                .where(NotificationRuleRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return NotificationRuleRow.created_at > subquery

        return inner


class NotificationRuleOrders:
    """Query orders for notification rules."""

    @staticmethod
    def name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return NotificationRuleRow.name.asc()
        return NotificationRuleRow.name.desc()

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return NotificationRuleRow.created_at.asc()
        return NotificationRuleRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return NotificationRuleRow.updated_at.asc()
        return NotificationRuleRow.updated_at.desc()
